import os
import glob
import shutil
import urllib.parse
import logging
import traceback
logger = logging.getLogger(__name__)

from appyter.ext.fs import Filesystem as FS


class Filesystem:
  def __init__(self, uri):
    self._uri = uri
    self._prefix = self._uri.path
  #
  def __enter__(self):
    return self
  #
  def path(self, path=''):
    return FS.join(self._prefix, path)
  #
  def close(self):
    pass
  #
  def open(self, path, mode='r'):
    try:
      assert path
      if mode[0] in {'w', 'a'}:
        os.makedirs(os.path.dirname(FS.join(self._prefix, path)), exist_ok=True)
      return open(FS.join(self._prefix, path), mode=mode)
    except FileNotFoundError:
      raise Exception(f"No such file or directory: {path}")
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to access {path}")
  #
  def exists(self, path):
    try:
      assert path
      return os.path.isfile(FS.join(self._prefix, path)) or os.path.islink(FS.join(self._prefix, path))
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to access {path}")
  #
  def ls(self, path=''):
    ls_path = FS.join(self._prefix, path) if path else self._prefix
    return [
      f[len(ls_path)+1:]
      for f in glob.glob(ls_path + '/*') + glob.glob(ls_path + '/**/*')
      if os.path.isfile(f)
    ]
  #
  def cp(self, src, dst):
    try:
      assert src and dst
      os.makedirs(os.path.dirname(FS.join(self._prefix, dst)), exist_ok=True)
      return shutil.copy(FS.join(self._prefix, src), FS.join(self._prefix, dst))
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to copy {src} to {dst}")
  #
  def link(self, src, dst):
    try:
      assert src and dst
      os.makedirs(os.path.dirname(FS.join(self._prefix, dst)), exist_ok=True)
      return os.link(FS.join(self._prefix, src), FS.join(self._prefix, dst))
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to link {src} to {dst}")
  #
  def rm(self, path, recursive=False):
    try:
      assert path
      if recursive:
        return shutil.rmtree(FS.join(self._prefix, path))
      else:
        return os.remove(FS.join(self._prefix, path))
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to access {path}")
  #
  def chmod_ro(self, path):
    try:
      assert path
      return os.chmod(FS.join(self._prefix, path), 400)
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to access {path}")
  #
  def __exit__(self, *args):
    pass



''' IPython magic for making templating easy~. This basically
just allows our jinja-type language to be executed in place
injecting the defaults into the environment so we can easily
debug the notebook at the same time as building the appyter.

The same call structure is used during preprocess and at runtime
but performing different tasks--this way setting up a notebook is
as simple as running it with different nbtemplate's being provided
for import.

Usage (put the following in the first cell):

```python
#%%appyter init
from appyter import magic
magic.init(lambda _=globals: _())
```
'''

'''
Setup given globals
'''
def init(_globals, verbose=False, ipynb='app.ipynb', mode='magic', **kwargs):
  ''' Initialize appyter magic.

  Sets up a jinj2 environment and injects %%appyter magic into your environment.
  
  :param _globals: (Dict[str, Any]) A callable with your globals for the purpose of injection, basically just: `lambda _=globals: _()`
  :param verbose: (Optional[bool]) Expand exception reporting to be more verbose
  '''
  import os
  import jinja2
  import jinja2.meta
  import traceback
  from appyter.context import get_env, get_jinja2_env
  env = get_jinja2_env(config=get_env(verbose=verbose, ipynb=ipynb, mode=mode, **kwargs))
  from IPython.core.magic import register_cell_magic
  from IPython.display import display, Markdown, HTML

  '''
  register_cell_magic allows function to execute an entire cell with the following call structure:
  ```python
  %%my_magic whatever
  all
  my
  data
  ```
  Results in a call:
  ```python
  my_magic(
    "whatever",
    """all
    my
    data"""
  )
  ```
  '''

  @register_cell_magic
  def appyter(line, cell):
    ''' Appyter Cell Magic: See Steps for more information.
    Compile jinja2 into source code, and then evaluate that
    source code.
    '''

    '''
    Step 1. Render cell with jinja2, removing empty lines.
    execute or display the results, modifying a copy of the
    current python globals dict.
    '''
    global_internal = _globals()
    cell_type = line.split('_')
    try:
      cell_lines = cell.splitlines()
      try:
        undeclared = jinja2.meta.find_undeclared_variables(env.parse(cell))
        if undeclared:
          for lineno, cell_line in enumerate(cell_lines):
            if any(v in cell_line for v in undeclared):
              display(HTML(f"<div style='color: red'>{lineno}: {cell_line}</div>"))
          raise Exception(f"undeclared variable(s) {undeclared}")
        template = env.from_string(cell)
        template_rendered = template.render()
      except Exception as e:
        if getattr(e, 'lineno', None) is not None:
          display(HTML(f"<div style='color: red'>{e.lineno}: {cell_lines[e.lineno-1]}</div>"))
        raise e
      #
      rendered_template_lines = list(filter(None, map(str.rstrip, template_rendered.splitlines())))
      try:
        if len(rendered_template_lines) > 0:
          if cell_type == ['markdown']:
            display(Markdown(template_rendered))
          elif 'code' in cell_type:
            rendered = '\n'.join(rendered_template_lines[:-1])
            rendered_last = rendered_template_lines[-1]
            display(Markdown('```python\n%s\n```' % ('\n'.join((rendered, rendered_last)))))
            #
            if 'eval' in cell_type:
              exec(rendered, global_internal)
              #
              try:
                display(eval(rendered_last, global_internal))
              except Exception as e:
                setattr(e, 'lineno', len(rendered_template_lines))
                raise e
            elif 'exec' in cell_type:
              exec('\n'.join((rendered, rendered_last)), global_internal)
          else:
            raise Exception('Unrecognized appyter cell_type')
        #
      except Exception as e:
        if getattr(e, 'lineno', None) is not None:
          display(HTML(f"<div style='color: red'>{e.lineno}: {rendered_template_lines[e.lineno-1]}</div>"))
        raise e
    except Exception as e:
      display(HTML(f"<div style='color: red'>Error: {e}</div>"))
      if verbose:
        traceback.print_exc()
      return
    '''
    Step 2. Check for new variables in the internal global
    and pass them to the python global scope. Check for
    new variables in the jinja2 template and pass them to the
    template environment global so they are available in
    future jinja2 calls.
    '''
    for k, v in global_internal.items():
      if not k.startswith('_'):
        _globals()[k] = v

    for k, v in template.module.__dict__.items():
      if not k.startswith('_'):
        env.globals[k] = v



import asyncio
import datetime
import functools
import importlib
from aiohttp import web
from collections import deque
from collections import OrderedDict, Counter
import logging
logger = logging.getLogger(__name__)

from appyter.orchestration.dispatcher.socketio import socketio

class LockedOrderedDict(OrderedDict):
  lock = asyncio.Lock()

# NOTE: This is questionable, this queue may run
#       on a different thread or process; in the case of a thread it's fine
#       because Queue is thread-safe, but not in the case of a process
#       we'd need a monkey-patched redis-Queue.
#

core = web.Application()
routes = web.RouteTableDef()

@routes.get('/')
async def on_get(request):
  async with core['tasks'].lock:
    return web.json_response(dict(
      tasks=list(core['tasks'].values()),
      ts=datetime.datetime.now().replace(
        tzinfo=datetime.timezone.utc
      ).isoformat(),
    ))

@routes.post('/')
async def on_submit(request):
  request_data = await request.json()
  async with core['tasks'].lock:
    core['tasks'][request_data['id']] = dict(request_data,
      debug=core['config']['DEBUG'],
      queued_ts=datetime.datetime.now().replace(
        tzinfo=datetime.timezone.utc
      ).isoformat(),
    )
  asyncio.create_task(core['dispatch_queue'].put(request_data['id']))
  return web.json_response(core['dispatch_queue'].qsize() + 1)

@core.on_startup.append
async def start_dispatchers(app):
  logger.info('Initializing dispatch...')
  #
  app['dispatch_queue'] = asyncio.Queue()
  app['tasks'] = LockedOrderedDict()
  #
  dispatch = functools.partial(
    importlib.import_module(
      '..dispatch.{}'.format(app['config']['DISPATCH']),
      __package__
    ).dispatch,
    debug=app['config']['DEBUG'],
    namespace=app['config']['KUBE_NAMESPACE'],
  )
  #
  logger.info('Starting background tasks...')
  for _ in range(app['config']['JOBS']):
    asyncio.create_task(
      dispatcher(
        queued=app['dispatch_queue'],
        tasks=app['tasks'],
        dispatch=dispatch,
        jobs_per_image=app['config']['JOBS_PER_IMAGE'],
      )
    )

core.add_routes(routes)

async def slow_put(queue, item):
  ''' Queue put debouncing
  '''
  await asyncio.sleep(1)
  await queue.put(item)

async def dispatcher(queued=None, tasks=None, dispatch=None, jobs_per_image=1):
  while True:
    job_id = await queued.get()
    async with tasks.lock:
      job = tasks[job_id]
      image_jobs = Counter(t.get('image') for t in tasks.values() if 'started_ts' in t).get(job.get('image'), 0)
      if image_jobs >= jobs_per_image:
        # TODO: push back onto a priority queue such that the next slot that opens uses this one
        #       currently, this appyter execution would end up on the back of the queue
        queued.task_done()
        asyncio.create_task(slow_put(queued, job_id))
        continue
      else:
        tasks[job['id']]['started_ts'] = datetime.datetime.now().replace(
          tzinfo=datetime.timezone.utc
        ).isoformat()
    #
    try:
      logger.info(f"Dispatching job {job_id}")
      await dispatch(job=job)
    except:
      import traceback
      logger.error(f"dispatch error: {traceback.format_exc()}")
    #
    async with tasks.lock:
      del tasks[job_id]
  #
  job.task_done()



import re
from appyter.fields import Field
from appyter.util import re_full

class TextField(Field):
  ''' Representing a field that accepts a multi-line string

  Be careful with this field, consider defining a constraint regex.

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param constraint: (Regex[str]) A regular expression for validating the file name.
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param default: (str) A default value as an example and for use during prototyping
  :param examples: (Optional[Dict[str, str]]) Named strings to provide as clickable examples
  :param hint: (Optional[str]) A hint to put in the field prior to content.
  :param rows: (Optional[int]) The number of rows (lines) in the textarea
  :param cols: (Optional[int]) The number of cols (horizontal characters) in the textarea
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Remaining arguments passed down to :class:`appyter.fields.Field`'s constructor.
  '''
  def __init__(self, constraint=r'.*', hint=None, **kwargs):
    super().__init__(
      constraint=constraint,
      hint=hint,
      **kwargs,
    )

  def constraint(self):
    if self.raw_value is None:
      return not self.args.get('required')
    else:
      return re.match(re_full(self.args['constraint']), self.raw_value, re.MULTILINE | re.DOTALL)

  @property
  def safe_value(self):
    # use repr to escape things, revert newlines
    v = repr(self.value).replace('\\r', '\r').replace('\\n', '\n')
    # convert single to triple quotes
    return v[0] + v[0] + v + v[0] + v[0]



import tempfile
import urllib.parse
import json
import re
import os
import time
import logging
logger = logging.getLogger(__name__)

from appyter.ext.fs.file import Filesystem as FSFilesystem


def sh(cmd):
  from subprocess import Popen, PIPE
  return Popen(cmd, stdout=PIPE)

def slugify(s):
  return re.sub(r'[^a-zA-Z0-9-]+', '', s)

def sync_async_sleep(amnt, asynchronous=False):
  if asynchronous:
    import asyncio
    asyncio.ensure_future(asyncio.sleep(amnt))
  else:
    time.sleep(amnt)

class RcloneParse:
  @staticmethod
  def get(scheme):
    return getattr(RcloneParse, '_{}'.format(scheme))

  @staticmethod
  def _s3(uri):
    config = dict(urllib.parse.parse_qsl(uri.query))
    if 'env_auth' not in config:
      config['env_auth'] = 'false'
    #
    if 'provider' not in config:
      if uri.hostname.endswith('s3.amazonaws.com'):
        config['provider'] = 'AWS'
      else:
        config['provider'] = 'Minio'
    #
    if 'endpoint' not in config:
      if config['provider'] != 'AWS':
        config['endpoint'] = f"{'https' if config.get('use_ssl') else 'http'}://{uri.hostname}"
        if uri.port:
          config['endpoint'] += f":{uri.port}"
    #
    if uri.username:
      config['access_key_id'] = uri.username
    if uri.password:
      config['secret_access_key'] = uri.password
    #
    return [el for key, value in config.items() for el in [key, value]]

class Filesystem(FSFilesystem):
  def __init__(self, uri, asynchronous=False):
    self._uri = uri
    self._scheme = '+'.join(scheme for scheme in self._uri.scheme.split('+') if scheme != 'rclone')
    self._remote = slugify(f"{self._scheme}:{self._uri.hostname}{self._uri.path}?{self._uri.query}")
    if self._remote not in json.load(sh(['rclone', 'config', 'dump']).stdout).keys():
      if sh(['rclone', 'config', 'create', self._remote, self._scheme] + RcloneParse.get(self._scheme)(self._uri)).wait() != 0:
        raise Exception('Error configuring rclone')
    self._tmpdir = tempfile.mkdtemp()
    logger.debug(f"Mounting {self._remote}:{self._uri.path[1:]} on {self._tmpdir}")
    self._mount = sh([
      'rclone', 'mount', '--vfs-cache-mode', 'writes', f"{self._remote}:{self._uri.path[1:]}", self._tmpdir,
    ])
    sync_async_sleep(0.1, asynchronous=asynchronous)
    while not os.path.ismount(self._tmpdir):
      if self._mount.poll() is not None:
        raise Exception(f"Mount exited with code {self._mount.returncode} before mounting")
      sync_async_sleep(0.1, asynchronous=asynchronous)
    super().__init__(urllib.parse.urlparse('file://' + self._tmpdir))
  #
  def __exit__(self, *args):
    self.close()
  #
  def link(self, src, dst):
    if self._scheme == 's3':
      logger.info('WARNING: s3 does not support links, copying')
      return self.cp(src, dst)
    return super().link(src, dst)
  #
  def close(self):
    logger.debug('Unmounting...')
    self._mount.terminate()
    self._mount.wait()
    os.rmdir(self._tmpdir)



''' This module enables you to turn your jupyter notebook into a jinja2 template-driven web application. Or just parse for other purposes.
'''
import os
from appyter.magic import init

__version__ = open(os.path.join(os.path.dirname(__file__), 'VERSION'), 'r').read().strip()



def s3_to_url(s3_uri):
  import urllib.parse
  uri = urllib.parse.urlparse(s3_uri)
  q = dict(urllib.parse.parse_qsl(uri.query))
  return urllib.parse.urlparse(f"{'https' if q.get('use_ssl') else 'http'}://{uri.hostname}:{uri.port or (443 if q.get('use_ssl') else 80)}/{uri.path}")

def serve(app_path, **kwargs):
  import os
  import sys
  import time
  import appyter
  import functools
  import logging
  logger = logging.getLogger(__name__)
  from subprocess import Popen
  from appyter.ext.fs import Filesystem
  from appyter.context import get_env, get_jinja2_env, find_blueprints, get_appyter_directory
  from appyter.util import join_routes
  from appyter.profiles.default.filters.url_for import url_for
  config = get_env(**kwargs)
  logger.info(kwargs)
  env = get_jinja2_env(config=config)
  with Filesystem('tmpfs://') as tmp_fs:
    logger.info(f"Working directory {tmp_fs.path()}")
    #
    logger.info(f"Pre-rendering pages...")
    with Filesystem(config['CWD']).open(config['IPYNB']) as fr:
      from appyter.parse.nb import nb_from_ipynb_io
      nbtemplate = nb_from_ipynb_io(fr)
    with tmp_fs.open('index.html', 'w') as fw:
      from appyter.render.form import render_form_from_nbtemplate
      fw.write(render_form_from_nbtemplate(env, nbtemplate))
    with tmp_fs.open('index.json', 'w') as fw:
      import json
      from appyter.render.nbinspect import render_nbtemplate_json_from_nbtemplate
      json.dump(render_nbtemplate_json_from_nbtemplate(env, nbtemplate), fw)
    with tmp_fs.open('landing.html', 'w') as fw:
      env.get_template('landing.j2').stream(
        _nb=os.path.basename(config['IPYNB']),
      ).dump(fw)
    #
    logger.info(f"Generating production config...")
    with tmp_fs.open('supervisord.conf', 'w') as fw:
      env.get_template('production/supervisord.conf.j2').stream(_tmp_fs=tmp_fs, sys=sys, str=str).dump(fw)
    with tmp_fs.open('nginx.conf', 'w') as fw:
      env.get_template('production/nginx.conf.j2').stream(
        _tmp_fs=tmp_fs, os=os,
        s3_to_url=s3_to_url,
        get_appyter_directory=get_appyter_directory,
        find_blueprints=find_blueprints,
      ).dump(fw)
    logger.info(f"Starting production instance at http://{config['HOST']}:{config['PORT']}{config['PREFIX']} ...")
    with Popen(['supervisord', '-n', '-c', tmp_fs.path('supervisord.conf')]) as proc:
      try:
        sys.exit(proc.wait())
      except KeyboardInterrupt:
        proc.terminate()
        sys.exit(proc.wait())



''' Launch job using docker directly
'''

import os
import sys
import json
import asyncio
import logging
logger = logging.getLogger(__name__)

async def dispatch(job=None, **kwargs):
  args = ['docker', 'run']
  args += ['--device', '/dev/fuse']
  args += ['--privileged']
  try:
    proc = await asyncio.create_subprocess_exec(*[
      'docker', 'inspect', os.environ['HOSTNAME'],
    ], stdout=asyncio.subprocess.PIPE)
    proc_stdout, _ = await proc.communicate()
    conf = json.loads(proc_stdout.decode())
    args += ['--network', conf[0]['HostConfig']['NetworkMode']]
  except:
    import traceback
    logger.warn(traceback.format_exc())
    args += []
  #
  args += [
    job['image'],
    'appyter', 'orchestration', 'job', json.dumps(job),
  ]
  proc = await asyncio.create_subprocess_exec(*args, stdout=sys.stdout, stderr=sys.stderr)
  return await proc.wait()



import socketio
import logging
logger = logging.getLogger(__name__)

from appyter.ext.socketio.priority_queued_emit import PriorityQueuedEmitMixin

class AsyncServer(PriorityQueuedEmitMixin, socketio.AsyncServer):
  ''' Server with packet forwarding
  '''
  klass = socketio.AsyncServer

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self._emit_enabled.set()
    #
    this = self
    @self.on('forward')
    async def forward(sid, data):
      logger.debug('Server forward')
      if data.get('to'):
        await this.emit(data['event'], data['data'], priority=max(0, data.get('priority', 0)) + 1, to=data['to'])
      elif data.get('room'):
        await this.emit(data['event'], data['data'], priority=max(0, data.get('priority', 0)) + 1, room=data['room'])



import click
import logging
logger = logging.getLogger(__name__)

from appyter.orchestration.cli import orchestration
from appyter.ext.click import click_option_setenv


def create_app(**kwargs):
  from aiohttp import web
  #
  from appyter.orchestration.dispatcher.core import core
  from appyter.orchestration.dispatcher.socketio import socketio
  from appyter.util import join_routes
  #
  logging.basicConfig(
    level=logging.DEBUG if kwargs.get('debug') else logging.WARNING,
    format='%(name)s %(message).80s',
  )
  #
  logger.info('Initializing aiohttp...')
  config = dict(
    HOST=kwargs.get('host'),
    PORT=kwargs.get('port'),
    PROXY=kwargs.get('proxy'),
    JOBS=kwargs.get('jobs'),
    JOBS_PER_IMAGE=kwargs.get('jobs_per_image'),
    DEBUG=kwargs.get('debug'),
    PREFIX=kwargs.get('prefix'),
    KUBE_NAMESPACE=kwargs.get('kube_namespace'),
    DISPATCH=kwargs.get('dispatch'),
  )
  if config['PREFIX'].rstrip('/'):
    app = web.Application()
    app['config'] = core['config'] = config
    #
    logger.info('Registering prefix redirect')
    async def redirect_to_prefix(request):
      path = request.match_info['path']
      raise web.HTTPFound(join_routes(app['config']['PREFIX'], path) + '/')
    app.router.add_get('/{path:[^/]*}', redirect_to_prefix)
    app.add_subapp(config['PREFIX'].rstrip('/'), core)
  else:
    app = core
    app['config'] = config
  #
  logger.info('Initializing socketio...')
  socketio.attach(app, join_routes(app['config']['PREFIX'], 'socket.io'))
  #
  return app

@orchestration.command(help='Run the appyter dispatcher')
@click_option_setenv('--prefix', envvar='APPYTER_PREFIX', default='/', help='Specify the prefix for which to mount the webserver onto')
@click_option_setenv('--host', envvar='APPYTER_HOST', default='127.0.0.1', help='The host the flask server should run on')
@click_option_setenv('--port', envvar='APPYTER_PORT', type=int, default=5000, help='The port this flask server should run on')
@click_option_setenv('--proxy', envvar='APPYTER_PROXY', type=bool, default=False, help='Whether this is running behind a proxy and the IP should be fixed for CORS')
@click_option_setenv('--jobs', envvar='APPYTER_JOBS', type=int, default=2, help='Number of concurrent jobs to dispatch')
@click_option_setenv('--jobs-per-image', envvar='APPYTER_JOBS_PER_IMAGE', type=int, default=1, help='Number of concurrent jobs to dispatch for any individual appyter image')
@click_option_setenv('--debug', envvar='APPYTER_DEBUG', type=bool, default=True, help='Whether or not we should be in debugging mode, not for use in multi-tenant situations')
@click_option_setenv('--kube-namespace', envvar='APPYTER_KUBE_NAMESPACE', type=str, default='default', help='The kubernetes namespace (kubernetes dispatch)')
@click_option_setenv('--dispatch', envvar='APPYTER_DISPATCH', type=str, default='native', help='The dispatcher mechanism to use (see list-dispatchers)')
def dispatcher(*args, **kwargs):
  from aiohttp import web
  from appyter.orchestration.dispatcher.socketio import socketio
  from appyter.util import dict_filter_none
  #
  app = create_app(**kwargs)
  web.run_app(app, host=app['config']['HOST'], port=int(app['config']['PORT']))



import os
import sys
import click
from appyter.cli import cli
from appyter.ext.click import click_option_setenv

@cli.command(
  help='A simple alias for jupyter notebook which re-uses appyter environment config',
  context_settings=dict(
    ignore_unknown_options=True,
  ),
)
@click_option_setenv('--data-dir', envvar='APPYTER_DATA_DIR', default=None, help='The directory to store data of executions')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_option_setenv('--host', envvar='APPYTER_HOST', default='0.0.0.0', help='The host to bind to')
@click_option_setenv('--port', envvar='APPYTER_PORT', type=int, default=5000, help='The port this server should run on')
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def serve(data_dir, cwd, host, port, args):
  import tempfile
  from subprocess import Popen
  # add cwd (the appyter itself) to the PYTHONPATH
  sys.path.insert(0, os.path.realpath(cwd))
  # create blank data_dir as tmpdir if it doesn't exist
  if data_dir is None:
    data_dir = tempfile.mkdtemp()
  #
  # but use data_dir (the appyter files) as the cwd of the jupyter notebook
  os.makedirs(data_dir, exist_ok=True)
  #
  # prepare default flags
  flags = {
    'NotebookApp.ip': host,
    'NotebookApp.port': port,
    'NotebookApp.token': '',
    'NotebookApp.password': '',
  }
  # look through args and replace any overrides
  extra_args = []
  for arg in args:
    if arg.startswith('--'):
      name, value = arg[2:].split('=', maxsplit=1)
      flags[name] = value
    else:
      extra_args.append(arg)
  #
  # run jupyter notebook interactively and exit with it
  proc = Popen(
    [
      'jupyter', 'notebook',
      *[
        f"--{key}={value}"
        for key, value in flags.items()
      ],
      *extra_args,
    ],
    stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr,
    cwd=os.path.realpath(data_dir),
    env=dict(
      PYTHONPATH=':'.join(sys.path),
      PATH=os.environ['PATH']
    ),
  )
  exit_code = None
  while exit_code is None:
    try:
      exit_code = proc.wait()
    except:
      pass
  #
  sys.exit(exit_code)



import os
import uuid
import json
import traceback
from flask import Blueprint, request, redirect, abort, send_file, url_for, current_app, jsonify, make_response
from werkzeug.exceptions import BadRequest

from appyter.context import get_jinja2_env
from appyter.ext.fs import Filesystem
from appyter.parse.nb import nb_from_ipynb_io, nb_to_ipynb_io
from appyter.util import secure_filepath, exception_as_dict
from appyter.render.form import render_form_from_nbtemplate
from appyter.render.nbconstruct import render_nb_from_nbtemplate
from appyter.render.nbinspect import render_nbtemplate_json_from_nbtemplate
from appyter.render.flask_app.download import upload_from_request
from appyter.render.flask_app.util import sanitize_sha1sum, sanitize_uuid, route_join_with_or_without_slash, collapse, sha1sum_io, sha1sum_dict


core = Blueprint('__main__', __name__)

_fields = None
def get_fields():
  ''' Helper to get/cache fields even if we're on a different thread
  '''
  global _fields
  if not _fields or current_app.config['DEBUG']:
    fs = Filesystem(current_app.config['CWD'])
    with fs.open(current_app.config['IPYNB'], 'r') as fr:
      env = get_jinja2_env(config=current_app.config)
      nbtemplate = nb_from_ipynb_io(fr)
      _fields = render_nbtemplate_json_from_nbtemplate(env, nbtemplate)
  return _fields

_ipynb_hash = None
def get_ipynb_hash():
  global _ipynb_hash
  if not _ipynb_hash or current_app.config['DEBUG']:
    fs = Filesystem(current_app.config['CWD'])
    _ipynb_hash = sha1sum_io(fs.open(current_app.config['IPYNB'], 'rb'))
  return _ipynb_hash

def prepare_data(data):
  # Assert that all fields are accounted for
  fields = {field['args']['name'] for field in get_fields()}
  unrecognized = set(data.keys()) - fields
  if unrecognized:
    raise BadRequest(f"Unrecognized fields: {', '.join(unrecognized)}")
  return data

def prepare_formdata(req):
  ''' Get data defined for each field from json/form. Also process POSTed files
  '''
  fields = get_fields()
  file_fields = {
    field['args']['name']
    for field in fields
    if field['field'] == 'FileField'
  }
  data = {}
  if req.form:
    data.update({
      field['args']['name']: collapse(req.form.getlist(field['args']['name']))
      for field in fields
      if field['args']['name'] in req.form
    })
    data.update(upload_from_request(req, file_fields))
  elif req.json:
    data.update({
      field['args']['name']: req.json.get(field['args']['name'])
      for field in fields
      if field['args']['name'] in req.json
    })
  return prepare_data(data)

def prepare_results(data):
  results_hash = sha1sum_dict(dict(ipynb=get_ipynb_hash(), data=data))
  data_fs = Filesystem(current_app.config['DATA_DIR'])
  results_path = Filesystem.join('output', results_hash)
  if not data_fs.exists(Filesystem.join(results_path, current_app.config['IPYNB'])):
    # prepare files to be linked and update field to use filename
    file_fields = {
      field['args']['name']
      for field in get_fields()
      if field['field'] == 'FileField'
    }
    links = []
    files = {}
    for file_field in file_fields:
      if fdata := data.get(file_field):
        content_hash, filename = fdata.split('/', maxsplit=1)
        content_hash = sanitize_sha1sum(content_hash)
        filename = secure_filepath(filename)
        links.append((
          Filesystem.join('input', content_hash),
          Filesystem.join(results_path, filename)
        ))
        files[filename] = filename
        data[file_field] = filename
    # construct notebook
    env = get_jinja2_env(config=current_app.config, context=data, session=results_hash)
    fs = Filesystem(current_app.config['CWD'])
    with fs.open(current_app.config['IPYNB'], 'r') as fr:
      nbtemplate = nb_from_ipynb_io(fr)
    # in case of constraint failures, we'll fail here
    nb = render_nb_from_nbtemplate(env, nbtemplate, files=files)
    # actually link all input files into output directory
    for src, dest in links:
      data_fs.link(src, dest)
    # write notebook
    nbfile = Filesystem.join(results_path, os.path.basename(current_app.config['IPYNB']))
    with data_fs.open(nbfile, 'w') as fw:
      nb_to_ipynb_io(nb, fw)
  #
  return results_hash

@route_join_with_or_without_slash(core, methods=['POST'])
def post_index():
  mimetype = request.accept_mimetypes.best_match([
    'text/html',
    'application/json',
  ], 'text/html')
  #
  try:
    data = prepare_formdata(request)
    result_hash = prepare_results(data)
    error = None
  except Exception as e:
    traceback.print_exc()
    error = exception_as_dict(e)
  #
  if mimetype in {'text/html'}:
    if error: abort(406)
    else: return redirect(url_for('__main__.data_files', path=result_hash + '/'), 303)
  elif mimetype in {'application/json'}:
    if error is not None:
      return make_response(jsonify(error=error), 406)
    else:
      return make_response(jsonify(session_id=result_hash), 200)
  else:
    abort(404)

@route_join_with_or_without_slash(core, 'ssr', methods=['POST'])
def post_ssr():
  env = get_jinja2_env(config=current_app.config)
  try:
    ctx = request.get_json()
    assert ctx['field'].endswith('Field'), 'Invalid field'
    return env.globals[ctx['field']](**ctx['args']).render()
  except Exception as e:
    return make_response(jsonify(error=exception_as_dict(e)), 406)



import json
from appyter.fields import Field

class BoolField(Field):
  ''' Represing a true or false value with a checkbox.

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param default: (bool) A default value as an example and for use during prototyping
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Additional keyword arguments used by other fields
  '''
  def __init__(self, **kwargs):
    super().__init__(**kwargs)

  @property
  def raw_value(self):
    if self.args['value'] in {'on', 'checked', 'yes', 'true', 'True', True}:
      return True
    elif self.args['value'] in {'off', 'unchecked', 'no', 'false', 'False', False}:
      return False
    elif self.args['value'] is None and not self.args.get('required'):
      return None
    else:
      raise Exception('Invalid value for bool field')

  @property
  def choices(self):
    return [True, False]



import os
import json
import click
import nbformat as nbf
from copy import deepcopy

from appyter import __version__
from appyter.cli import cli
from appyter.ext.fs import Filesystem
from appyter.context import get_env, get_jinja2_env
from appyter.parse.nb import nb_from_ipynb_io, nb_to_ipynb_io
from appyter.parse.nbtemplate import cell_match
from appyter.ext.click import click_option_setenv, click_argument_setenv

def render_cell(env, cell):
  ''' Render a single cell, calling jinja2 templates when necessary
  (only on %%appyter -matched cells) and converting cells
  to markdown / removing them as necessary.
  '''
  if cell.cell_type == 'code':
    cell.outputs = []
    cell['execution_count'] = None

  cell_m = cell_match.match(cell.source)
  if cell_m:
    cell_type = cell_m.group('type').strip().split('_')
    cell_source = cell_m.group('source').strip()

    if cell_type == ['init']:
      return None

    template = env.from_string(cell_source)
    rendered = template.render()
    trimmed = '\n'.join(
      line
      for line in rendered.splitlines()
      if line.strip() != ''
    )

    # Variables added via jinja should get added to globals
    #  for subsequent evaluation
    for k, v in template.module.__dict__.items():
      if not k.startswith('_'):
        env.globals[k] = v

    # Alter cell render based on cell_type
    if cell_type == ['markdown']:
      cell = nbf.v4.new_markdown_cell(rendered)
    elif 'hide' in cell_type or not trimmed:
      cell = None
    else:
      cell = nbf.v4.new_code_cell(trimmed)

  return cell


def render_nb_from_nbtemplate(env, nb, files={}):
  ''' Render the notebook by rendering the jinja2 templates using the context in env.
  '''
  nb = deepcopy(nb)
  nb.cells = list(filter(None, [
    render_cell(
      env,
      cell,
    )
    for cell in nb.cells
  ]))
  if 'appyter' not in nb.metadata:
    nb.metadata['appyter'] = {}
  nb.metadata['appyter']['nbconstruct'] = dict(
    version=__version__,
    filename=env.globals['_config']['IPYNB'],
    files=files,
  )
  return nb

@cli.command(help='Construct jupyter notebook from appyter and arguments')
@click.option('-i', '--context', envvar='APPYTER_CONTEXT', default='-', type=click.File('r'), help='JSON serialized context mapping field names to values')
@click.option('-o', '--output', envvar='APPYTER_OUTPUT', default='-', type=click.File('w'), help='The output location of the serialized jupyter notebook')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def nbconstruct(cwd, ipynb, context, output, **kwargs):
  context = json.load(context)
  env = get_jinja2_env(
    config=get_env(cwd=cwd, ipynb=ipynb, mode='construct', **kwargs),
    context=context,
  )
  nbtemplate = nb_from_ipynb_io(Filesystem(cwd).open(ipynb, 'r'))
  nb = render_nb_from_nbtemplate(env, nbtemplate)
  nb_to_ipynb_io(nb, output)



import aiohttp
import traceback
import logging
logger = logging.getLogger(__name__)

from appyter.ext.fs import Filesystem
from appyter.render.flask_app.socketio import socketio
from appyter.render.flask_app.util import secure_filepath, secure_url, sha1sum_io, generate_uuid

# organize file by content hash
def organize_file_content(data_fs, tmp_fs, tmp_path):
  with tmp_fs.open(tmp_path, 'rb') as fr:
    content_hash = sha1sum_io(fr)
  data_path = Filesystem.join('input', content_hash)
  if not data_fs.exists(data_path):
    Filesystem.mv(src_fs=tmp_fs, src_path=tmp_path, dst_fs=data_fs, dst_path=data_path)
    data_fs.chmod_ro(data_path)
  return content_hash

# download from remote
async def download_with_progress_and_hash(sid, data_fs, name, url, path, filename):
  # TODO: worry about files that are too big/long
  with Filesystem('tmpfs://') as tmp_fs:
    await socketio.emit('download_start', dict(name=name, filename=filename), room=sid)
    try:
      async with aiohttp.ClientSession() as client:
        async with client.get(url) as resp:
          # NOTE: this may become an issue if ever someone wants actual html
          assert resp.content_type != 'text/html', 'Expected data, got html'
          resp.headers.get('Content-Length', -1)
          chunk = 0
          chunk_size = 1024*8
          total_size = resp.headers.get('Content-Length', -1)
          async def reporthook(chunk):
            await socketio.emit(
              'download_progress',
              dict(name=name, chunk=chunk, chunk_size=chunk_size, total_size=total_size),
              room=sid,
            )
          with tmp_fs.open(path, 'wb') as fw:
            await reporthook(chunk)
            while buf := await resp.content.read(chunk_size):
              fw.write(buf)
              chunk += 1
              await reporthook(chunk)
    except Exception as e:
      logger.error(f"download error: {traceback.format_exc()}")
      await socketio.emit(
        'download_error',
        dict(name=name, filename=filename, url=url, error=str(e)),
        room=sid,
      )
    else:
      await socketio.emit(
        'download_complete',
        dict(
          name=name, filename=filename,
          full_filename='/'.join((organize_file_content(data_fs, tmp_fs, path), filename)),
        ),
        room=sid,
      )

@socketio.on('download_start')
async def download(sid, data):
  async with socketio.session(sid) as sess:
    config = sess['config']
  #
  data_fs = Filesystem(config['DATA_DIR'])
  name = data.get('name')
  # TODO: hash based on url?
  # TODO: s3 bypass
  url = secure_url(data.get('url'))
  filename = secure_filepath(data.get('file'))
  await socketio.emit('download_queued', dict(name=name, filename=filename), room=sid)
  await download_with_progress_and_hash(
    sid=sid,
    data_fs=data_fs,
    name=name,
    url=url,
    path=generate_uuid(),
    filename=filename,
  )


# upload from client
@socketio.on("siofu_start")
async def siofu_start(sid, data):
  try:
    path = generate_uuid()
    filename = secure_filepath(data.get('name'))
    tmp_fs = Filesystem('tmpfs://')
    async with socketio.session(sid) as sess:
      sess['file_%d' % (data.get('id'))] = dict(
        data,
        path=path,
        name=filename,
        bytesLoaded=0,
        tmp_fs=tmp_fs,
        fh=tmp_fs.open(path, 'wb'),
      )
    await socketio.emit(
      'siofu_ready',
      {
        'id': data.get('id'),
        'name': None,
      },
      room=sid,
    )
  except Exception as e:
    logger.error(traceback.format_exc())
    await socketio.emit('siofu_error', str(e), room=sid)

@socketio.on("siofu_progress")
async def siofu_progress(sid, evt):
  async with socketio.session(sid) as sess:
    sess['file_%d' % (evt['id'])]['fh'].write(evt['content'])
  logger.debug(f"progress: {evt}")
  await socketio.emit("siofu_chunk", dict(
    id=evt['id'],
  ), room=sid)

@socketio.on("siofu_done")
async def siofu_done(sid, evt):
  async with socketio.session(sid) as sess:
    sess['file_%d' % (evt['id'])]['fh'].close()
    config = sess['config']
    data_fs = Filesystem(config['DATA_DIR'])
    tmp_fs = sess['file_%d' % (evt['id'])]['tmp_fs']
    path = sess['file_%d' % (evt['id'])]['path']
    filename = sess['file_%d' % (evt['id'])]['name']
    content_hash = organize_file_content(data_fs, tmp_fs, path)
    tmp_fs.close()
    del sess['file_%d' % (evt['id'])]
  #
  await socketio.emit('siofu_complete', dict(
    id=evt['id'],
    detail=dict(
      full_filename='/'.join((content_hash, filename)),
    )
  ), room=sid)

# upload from client with POST
def upload_from_request(req, fnames):
  from flask import current_app
  data_fs = Filesystem(current_app.config['DATA_DIR'])
  data = dict()
  for fname in fnames:
    fh = req.files.get(fname)
    if fh:
      filename = secure_filepath(fh.filename)
      path = generate_uuid()
      with Filesystem('tmpfs://') as tmp_fs:
        with tmp_fs.open(path, 'wb') as fw:
          fh.save(fw)
        data[fname] = '/'.join((organize_file_content(data_fs, tmp_fs, path), filename))
  return data



''' Represent the areas that can be handled externally in production
'''
import os
from flask import request, current_app, send_file, send_from_directory, abort, jsonify
from werkzeug.exceptions import NotFound

from appyter.ext.fs import Filesystem
from appyter.render.flask_app.core import core
from appyter.render.flask_app.util import route_join_with_or_without_slash
from appyter.context import get_appyter_directory
from appyter.parse.nb import nb_from_ipynb_io
from appyter.render.form import render_form_from_nbtemplate
from appyter.render.nbinspect import render_nbtemplate_json_from_nbtemplate
from appyter.context import get_jinja2_env


@route_join_with_or_without_slash(core, methods=['GET'])
def get_index():
  mimetype = request.accept_mimetypes.best_match([
    'text/html',
    'application/json',
    'application/vnd.jupyter',
  ], 'text/html')
  fs = Filesystem(current_app.config['CWD'])
  if mimetype in {'text/html'}:
    with fs.open(current_app.config['IPYNB'], 'r') as fr:
      env = get_jinja2_env(config=current_app.config)
      nbtemplate = nb_from_ipynb_io(fr)
    return render_form_from_nbtemplate(env, nbtemplate)
  elif mimetype in {'application/json'}:
    with fs.open(current_app.config['IPYNB'], 'r') as fr:
      env = get_jinja2_env(config=current_app.config)
      nbtemplate = nb_from_ipynb_io(fr)
    return jsonify(render_nbtemplate_json_from_nbtemplate(env, nbtemplate))
  elif mimetype in {'application/vnd.jupyter'}:
    return send_file(fs.open(current_app.config['IPYNB'], 'rb'), attachment_filename=current_app.config['IPYNB'], mimetype=mimetype)
  else:
    abort(404)

@core.route('/favicon.ico', methods=['GET'])
def favicon():
  static = Filesystem(current_app.config['STATIC_DIR'])
  if static.exists('favicon.ico'):
    return send_file(static.open('favicon.ico', 'rb'), attachment_filename='favicon.ico')
  abort(404)

@core.route('/static/<path:filename>', methods=['GET'])
def static(filename):
  static = Filesystem(current_app.config['STATIC_DIR'])
  if static.exists(filename):
    return send_file(static.open(filename, 'rb'), attachment_filename=filename)
  #
  try:
    return send_from_directory(get_appyter_directory(f"profiles/{current_app.config['PROFILE']}/static"), path=filename)
  except NotFound:
    return send_from_directory(get_appyter_directory('profiles/default/static'), path=filename)

@route_join_with_or_without_slash(core, '<path:path>', methods=['GET'])
def data_files(path):
  if path.endswith('/'):
    mimetype = request.accept_mimetypes.best_match([
      'text/html',
      'application/json',
      'application/vnd.jupyter',
    ], 'text/html')
    if mimetype == 'text/html':
      fs = Filesystem(current_app.config['CWD'])
      env = get_jinja2_env(config=current_app.config)
      return env.get_template('landing.j2').render(
        _nb=os.path.basename(current_app.config['IPYNB']),
      )
    else:
      data_fs = Filesystem(Filesystem.join(current_app.config['DATA_DIR'], 'output'))
      path += current_app.config['IPYNB']
      if data_fs.exists(path):
        return send_file(data_fs.open(path, 'rb'), attachment_filename=os.path.basename(path))
  else:
    data_fs = Filesystem(Filesystem.join(current_app.config['DATA_DIR'], 'output'))
    if data_fs.exists(path):
      return send_file(data_fs.open(path, 'rb'), attachment_filename=os.path.basename(path))
  abort(404)



import shutil
import tempfile
import urllib.parse
from appyter.ext.fs.file import Filesystem as FSFilesystem

class Filesystem(FSFilesystem):
  def __init__(self, uri):
    self._tmpdir = tempfile.mkdtemp(dir=uri.path) if uri.path else tempfile.mkdtemp()
    super().__init__(urllib.parse.urlparse('file://' + self._tmpdir))
  #
  def __exit__(self, *args):
    self.close()
  #
  def close(self):
    shutil.rmtree(self._uri.path)



import os
import shutil
import importlib
import urllib.parse
from werkzeug.security import safe_join

class Filesystem:
  def __init__(self, uri, with_path=False, asynchronous=False):
    if '://' not in uri:
      uri = 'file://' + uri
    uri_parsed = urllib.parse.urlparse(uri)
    self._fs = None
    if uri_parsed.scheme == 'file':
      from appyter.ext.fs.file import Filesystem as FSFilesystem
      self._fs = FSFilesystem(uri_parsed)
    if uri_parsed.scheme == 'tmpfs':
      from appyter.ext.fs.tmpfs import Filesystem as TmpFilesystem
      self._fs = TmpFilesystem(uri_parsed)
    if uri_parsed.scheme == 's3':
      if not with_path:
        from appyter.ext.fs.s3 import Filesystem as S3Filesystem
        self._fs = S3Filesystem(uri_parsed, asynchronous=asynchronous)
      else:
        # we will need rclone to do s3 mounting with_path
        uri_parsed = uri_parsed._replace(scheme='s3+rclone')
    if 'rclone' in uri_parsed.scheme.split('+'):
      from appyter.ext.fs.rclone import Filesystem as RcloneFilesystem
      self._fs = RcloneFilesystem(uri_parsed, asynchronous=asynchronous)
    #
    if self._fs is None:
      raise Exception(f"Unrecognized scheme {uri}")
  #
  def __enter__(self):
    return self._fs.__enter__()
  #
  def path(self, path=''):
    return self._fs.path(path)
  #
  def close(self):
    return self._fs.close()
  #
  def open(self, path, mode='r'):
    return self._fs.open(path, mode=mode)
  #
  def exists(self, path):
    return self._fs.exists(path)
  #
  def ls(self, path=''):
    return self._fs.ls(path=path)
  #
  def cp(self, src, dst):
    return self._fs.cp(src, dst)
  #
  def link(self, src, dst):
    return self._fs.link(src, dst)
  #
  def rm(self, path, recursive=False):
    return self._fs.rm(path, recursive=recursive)
  #
  def chmod_ro(self, path):
    return self._fs.chmod_ro(path)
  #
  def __exit__(self, *args):
    return self._fs.__exit__(*args)
  #
  @staticmethod
  def join(*args):
    return safe_join(*args)
  #
  @staticmethod
  def cp(src_fs=None, src_path=None, dst_fs=None, dst_path=None):
    if src_fs == dst_fs:
      src_fs.cp(src_path, dst_path)
    else:
      with src_fs.open(src_path, 'rb') as fr:
        with dst_fs.open(dst_path, 'wb') as fw:
          shutil.copyfileobj(fr, fw)
  #
  @staticmethod
  def mv(src_fs=None, src_path=None, dst_fs=None, dst_path=None):
    if src_fs == dst_fs:
      src_fs.mv(src_path, dst_path)
    else:
      Filesystem.cp(src_fs=src_fs, src_path=src_path, dst_fs=dst_fs, dst_path=dst_path)
      src_fs.rm(src_path)
  #



import click
from appyter.ext.json import try_json_loads, try_json_dumps

class Json(click.ParamType):
  name = 'Json'

  def convert(self, value, param, ctx):
    return try_json_loads(value)

  def split_envvar_value(self, rv):
    v = try_json_loads(rv)
    return v if type(v) == list else [v]

def click_option_setenv(spec, envvar=None, **kwargs):
  ''' Like click.option but explicitly set os.environ as well.
  '''
  import os, re, functools
  m = re.match(r'^--(.+)$', spec)
  assert m
  var = m.group(1).replace('-', '_')
  def decorator(func):
    @click.option(spec, envvar=envvar, **kwargs)
    @functools.wraps(func)
    def wrapper(**kwargs):
      if kwargs.get(var) is not None:
        os.environ[envvar] = try_json_dumps(kwargs[var])
      return func(**kwargs)
    return wrapper
  return decorator

def click_argument_setenv(var, envvar=None, **kwargs):
  ''' Like click.argument but explicitly set os.environ as well.
  '''
  import os, re, functools
  def decorator(func):
    @click.argument(var, envvar=envvar, **kwargs)
    @functools.wraps(func)
    def wrapper(**kwargs):
      os.environ[envvar] = try_json_dumps(kwargs[var.replace('-', '_')])
      return func(**kwargs)
    return wrapper
  return decorator



from appyter.fields import Field

class ChoiceField(Field):
  ''' Represing a choice of string values with a combo box.

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param choices: (Union[List[str], Set[str], Dict[str, str]]) A set of choices that are available for this field or lookup table mapping from choice label to resulting value
  :param default: (str) A default value as an example and for use during prototyping
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Additional keyword arguments used by other fields
  '''
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    assert self.choices, 'ChoiceField requires choices'

  def constraint(self):
    if self.raw_value is None:
      return not self.args.get('required')
    else:
      return self.raw_value in self.choices

  @property
  def safe_value(self):
    return self.value



import json
import hashlib
import asyncio
import socketio
import logging
logger = logging.getLogger(__name__)

from appyter.ext.socketio.priority_queued_emit import PriorityQueuedEmitMixin

class AsyncClient(PriorityQueuedEmitMixin, socketio.AsyncClient):
  ''' Client with packet forwarding, emit buffering & chunking
  '''
  klass = socketio.AsyncClient
  CHUNK_SIZE = 1<<20 # 1 kb

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.connection_tracker()

  async def disconnect(self):
    logger.debug('Disconnecting...')
    if self._emit_enabled.is_set():
      logger.debug('Ensuring emit queue has been fully processed...')
      await self._emit_queue.join()
    logger.debug('Disconnect')
    await self.klass.disconnect(self)

  def connection_tracker(self):
    @self.event
    async def connect():
      self._emit_enabled.set()
    @self.event
    async def disconnect():
      self._emit_enabled.clear()

  async def _emit(self, evt, data, to=None, room=None, priority=0):
    if to is not None or room is not None:
      logger.debug('Client forwarding')
      await self._emit_queue.put((
        priority,
        next(self._emit_counter), 
        (
          'forward',
          dict(event=evt, data=data, to=to, room=room, priority=priority)
        ),
        dict()
      ))
    else:
      return await PriorityQueuedEmitMixin._emit(self, evt, data, to=to, room=room, priority=priority)

  async def emit(self, evt, data, priority=0, **kwargs):
    logger.debug('Client emit')
    serialized = json.dumps(data, sort_keys=True, separators=None, ensure_ascii=True)
    if len(serialized) > AsyncClient.CHUNK_SIZE:
      logger.debug(f"Large packet ({len(serialized)}), chunking...")
      sha1 = hashlib.sha1()
      sha1.update(serialized.encode())
      data_hash = sha1.hexdigest()
      n_chunks = len(serialized) // AsyncClient.CHUNK_SIZE + 1
      for n in range(0, n_chunks):
        await PriorityQueuedEmitMixin.emit(self,
          'chunked',
          dict(
            id=data_hash, event=evt, chunk=n, total=n_chunks,
            data=serialized[n*AsyncClient.CHUNK_SIZE:n*AsyncClient.CHUNK_SIZE+AsyncClient.CHUNK_SIZE]
          ),
          priority=priority+1,
          **kwargs,
        )
    else:
      await PriorityQueuedEmitMixin.emit(self, evt, data, priority=priority, **kwargs)



''' ```eval_rst
Filters represent classes derived from :class:`appyters.fields.Field`
 implementing more specific field types and are rendered in
 :mod:`appyters.profiles.default.templates`. The various other profiles
 can be used to extend or override those definitions.
``` '''


''' Launch kubernetes job from within a kubernetes cluster
'''

import json
import logging
logger = logging.getLogger(__name__)

from appyter.util import run_in_executor


def endless_watch(*args, **kwargs):
  from kubernetes import watch
  w = watch.Watch()
  s = iter(w.stream(*args, **kwargs))
  v = None
  while True:
    try:
      event = next(s)
      v = event['object'].metadata.resource_version
      yield event
    except StopIteration:
      s = iter(w.stream(*args, **kwargs, resource_version=v))

@run_in_executor
def dispatch(job=None, namespace='default', debug=False, **kwargs):
  logger.info(f"starting {job['id']}")
  from kubernetes import client, config
  config.load_incluster_config()
  coreV1 = client.CoreV1Api()
  batchV1 = client.BatchV1Api()
  batchV1.create_namespaced_job(
    namespace=namespace,
    body=client.V1Job(
      api_version='batch/v1',
      kind='Job',
      metadata=client.V1ObjectMeta(
        name=job['id'],
        annotations={
          f"container.apparmor.security.beta.kubernetes.io/appyter-{job['id']}": 'unconfined'
        },
      ),
      spec=client.V1JobSpec(
        template=client.V1PodTemplateSpec(
          spec=client.V1PodSpec(
            restart_policy='Never',
            containers=[
              client.V1Container(
                name=f"appyter-{job['id']}",
                image=job['image'],
                command=['appyter', 'orchestration', 'job', json.dumps(job)],
                security_context=client.V1SecurityContext(
                  privileged=True,
                  capabilities=client.V1Capabilities(
                    add=['SYS_ADMIN'],
                  ),
                ),
                volume_mounts=[
                  client.V1VolumeMount(
                    name='fuse',
                    mount_path='/dev/fuse',
                  )
                ],
              ),
            ],
            volumes=[
              client.V1Volume(
                name='fuse',
                host_path=client.V1HostPathVolumeSource(
                  path='/dev/fuse',
                )
              ),
            ],
          ),
        ),
        backoff_limit=1,
      ),
    ),
  )
  #
  for event in endless_watch(batchV1.list_namespaced_job, namespace, 
    label_selector=f"job-name={job['id']}"
  ):
    logger.debug(str(event))
    event_type = event['type']
    event_job = event['object']
    if event_type == 'MODIFIED':
      if event_job.status.succeeded or event_job.status.failed:
        break
  #
  if not debug:
    batchV1.delete_namespaced_job(job['id'], namespace)
    # delete associated pod(s)
    for event_pod in coreV1.list_namespaced_pod(namespace, label_selector=f"job-name={job['id']}", watch=False).items:
      coreV1.delete_namespaced_pod(event_pod.metadata.name, namespace)
  #
  logger.info(f"{job['id']} completed")



import os
import sys
import click
import traceback
from appyter import __version__
from appyter.cli import cli
from appyter.ext.click import click_option_setenv, click_argument_setenv
from appyter.util import join_routes

@cli.command(
  help='An alias for jupyter notebook which pre-fetches notebook data from a remote appyter and then launches serve',
  context_settings=dict(
    ignore_unknown_options=True,
  ),
)
@click_option_setenv('--data-dir', envvar='APPYTER_DATA_DIR', default=None, help='The directory to store data of executions')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_option_setenv('--host', envvar='APPYTER_HOST', default='0.0.0.0', help='The host to bind to')
@click_option_setenv('--port', envvar='APPYTER_PORT', type=int, default=5000, help='The port this server should run on')
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
@click_argument_setenv('uri', envvar='APPYTER_URI', nargs=1, type=str)
@click.pass_context
def fetch_and_serve(ctx, data_dir, cwd, host, port, args, uri):
  import json
  import tempfile
  import urllib.request
  import nbformat as nbf
  # fetch the actual notebook
  try:
    with urllib.request.urlopen(
      urllib.request.Request(uri, headers={ 'Accept': 'application/vnd.jupyter' })
    ) as fr:
      nb = nbf.read(fr, as_version=4)
  except:
    traceback.print_exc()
    click.echo('Error fetching appyter instance, is the url right?')
  #
  metadata = nb.get('metadata', {}).get('appyter', {})
  #
  nbconstruct_version = metadata.get('nbconstruct', {}).get('version', 'unknown')
  if nbconstruct_version and nbconstruct_version == __version__:
    pass
  else:
    click.echo(f"WARNING: this appyter was not created with this version, instance version was {nbconstruct_version} and our version is {__version__}. Proceed with caution")
  #
  nbexecute_version = metadata.get('nbexecute', {}).get('version', 'unknown')
  if nbexecute_version and nbexecute_version == __version__:
    pass
  else:
    click.echo(f"WARNING: this appyter was not executed with this version, instance version was {nbexecute_version} and our version is {__version__}. Proceed with caution")
  #
  if 'nbexecute' not in metadata:
    click.echo('WARNING: this appyter instance has not been executed, no results will be available')
  elif 'started' in metadata['nbexecute'] and 'completed' not in metadata['nbexecute']:
    click.echo('WARNING: this appyter is not finished running, no results will be available')
  elif 'started' in metadata['nbexecute'] and 'completed' in metadata['nbexecute']:
    click.echo(f"Appyter ran from {metadata['nbexecute']['started']} to {metadata['nbexecute']['completed']}")
  else:
    click.echo('WARNING: this appyter seems old, this may not work properly, please contact us and we can update it')
  # if tmpdir doesn't exist, create it
  if data_dir is None:
    data_dir = tempfile.mkdtemp()
  # write notebook to data_dir
  os.makedirs(data_dir, exist_ok=True)
  filename = metadata.get('nbconstruct', {}).get('filename', 'appyter.ipynb')
  with open(os.path.join(data_dir, filename), 'w') as fw:
    nbf.write(nb, fw)
  #
  # download all files to data_dir (get files from nbexecute if available otherwise fall back to nbconstruct input-files)
  files = metadata.get('nbexecute', {}).get('files', metadata.get('nbconstruct', {}).get('files', {}))
  for file, fileurl in files.items():
    # relative file paths (those without schemes) are relative to the base uri
    # TODO: in the future we might be able to mount these paths as we do in the job
    if '://' not in fileurl:
      fileurl = join_routes(uri, fileurl)[1:]
    click.echo(f"Fetching {file} from {fileurl}...")
    try:
      urllib.request.urlretrieve(fileurl, os.path.join(data_dir, file))
    except:
      traceback.print_exc()
      click.echo(f"WARNING: Error fetching '{file}' exported by the appyter")
  #
  click.echo(f"Done. Starting `appyter serve`...")
  # serve the bundle in jupyter notebook
  from appyter.helpers.serve import serve
  ctx.invoke(serve, cwd=cwd, data_dir=data_dir, host=host, port=port, args=[*args, filename])



from appyter.fields import Field

class DescriptionField(Field):
  ''' Representing text between fields

  Usage:
  ```python
  {% do DescriptionField(
    name='my_fieldname',
    text='My Description',
  ) %}
  ```

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param text: (str) A human readable, HTML parsable text
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param \**kwargs: Additional keyword arguments used by other fields
  '''
  def __init__(self, **kwargs):
    super().__init__(**kwargs)

  @property
  def raw_value(self):
    raise Exception('DescriptionField has no value')



from appyter.parse.nbtemplate import parse_fields_from_nbtemplate

def render_form_from_nbtemplate(env, nb):
  ''' Render a form with the Fields requested throughout the notebook.
  `form.j2` is responsible for calling render on each relevant field.
  '''
  return env.get_template('form.j2').render(
    _fields=parse_fields_from_nbtemplate(env, nb)
  )



import s3fs
import urllib.parse
import traceback
import logging
logger = logging.getLogger(__name__)

from appyter.ext.fs import Filesystem as FS

class Filesystem:
  def __init__(self, uri, asynchronous=False):
    try:
      self._uri = uri
      self._config = {'asynchronous': asynchronous}
      #
      if self._uri.username and self._uri.password:
        self._config['key'] = self._uri.username
        self._config['secret'] = self._uri.password
      elif self._uri.username:
        self._config['token'] = self._uri.username
      else:
        self._config.update(dict(urllib.parse.parse_qsl(self._uri.query)))
      self._config['client_kwargs'] = dict(endpoint_url=f"{'https' if self._config.get('use_ssl') else 'http'}://{self._uri.hostname}:{self._uri.port or (443 if self._config.get('use_ssl') else 80)}")
      self._config['config_kwargs'] = dict(signature_version='s3v4')
      #
      self._prefix = self._uri.path.lstrip('/').rstrip('/') + '/'
      self._fs = s3fs.S3FileSystem(**self._config)
      if asynchronous:
        import asyncio
        asyncio.ensure_future(self._fs._connect())
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An internal error occurred")
  #
  def path(self, path=''):
    return None
  #
  def __enter__(self):
    return self
  #
  def __exit__(self, *args):
    pass
  #
  def close(self):
    pass
  #
  def open(self, path, mode='r'):
    try:
      assert path
      return self._fs.open(self._prefix + path, mode=mode)
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to access {path}")
  #
  def exists(self, path):
    try:
      assert path
      return self._fs.exists(self._prefix + path)
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to access {path}")
  #
  def ls(self, path=''):
    ls_path = FS.join(self._prefix, path) if path else self._prefix
    return [
      f[len(ls_path)+1:]
      for f in self._fs.glob(ls_path + '/*') + self._fs.glob(ls_path + '/**/*')
      # note S3 only reports files anyway
    ]
  #
  def cp(self, src, dst):
    try:
      assert src and dst
      return self._fs.copy(self._prefix + src, self._prefix + dst)
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to copy {src} to {dst}")
  #
  def link(self, src, dst):
    logger.info('WARNING: s3 does not support links, copying')
    return self.cp(src, dst)
  #
  def rm(self, path, recursive=False):
    try:
      assert path
      return self._fs.rm(self._prefix + path, recursive=recursive)
    except Exception:
      logger.error(traceback.format_exc())
      raise Exception(f"An error occurred while trying to access {path}")
  #
  def chmod_ro(self, path):
    logger.info('WARNING: s3 does not support chmod')



import os
import aiohttp
import asyncio
import traceback
from subprocess import PIPE
import logging
logger = logging.getLogger(__name__)

from appyter.render.flask_app.core import prepare_data, prepare_results
from appyter.render.flask_app.socketio import socketio
from appyter.render.flask_app.util import sanitize_sha1sum, generate_uuid, join_routes
from appyter.ext.fs import Filesystem

# construct/join a notebook
@socketio.on('submit')
async def submit(sid, data):
  async with socketio.session(sid) as sess:
    config = sess['config']
    request_url = sess['request_url']
  #
  if type(data) == dict:
    data = prepare_data(data)
    result_hash = prepare_results(data)
  elif type(data) == str:
    result_hash = sanitize_sha1sum(data)
    assert result_hash is not None
  else:
    raise Exception('Unrecognized data type')
  #
  socketio.enter_room(sid, result_hash)
  await socketio.emit('status', 'Queuing execution', to=sid)
  job = dict(
    cwd=Filesystem.join(config['DATA_DIR'], 'output', result_hash),
    ipynb=os.path.basename(config['IPYNB']),
    session=result_hash,
    id=generate_uuid(),
  )
  #
  if config['DISPATCHER_URL']:
    job['url'] = join_routes(config['DISPATCHER_URL'], 'socket.io').lstrip('/') + '/'
  else:
    job['url'] = request_url
  #
  if config['DISPATCHER_IMAGE']:
    job['image'] = config['DISPATCHER_IMAGE']
  #
  if config['DISPATCHER']:
    queued = False
    backoff = 1
    while not queued:
      try:
        async with aiohttp.ClientSession(headers={'Content-Type': 'application/json'}) as client:
          async with client.post(config['DISPATCHER'], json=job) as resp:
            queue_size = await resp.json()
            await socketio.emit('status', f"Queued successfully, you are at position {queue_size} in the queue", to=sid)
            queued = True
      except asyncio.CancelledError:
        raise
      except Exception:
        logger.error(traceback.format_exc())
        if backoff < 60:
          await socketio.emit('status', f"Failed to contact orchestrator, trying again in {backoff}s...", to=sid)
          asyncio.sleep(backoff)
          backoff *= 2
        else:
          await socketio.emit('error', 'Failed to contact orchestrator, please try again later.', to=sid)
          break
  else:
    from appyter.orchestration.job.job import execute_async
    asyncio.create_task(execute_async(job))

@socketio.on('join')
async def _(sid, data):
  socketio.enter_room(sid, data)
  await socketio.emit('joined', dict(id=sid, session=data), room=data)



import uuid
import urllib.parse

from appyter.util import join_routes, safe_join, secure_url, secure_filepath

def sha1sum_io(io, chunk_size=65536):
  import hashlib
  sha1 = hashlib.sha1()
  while buf := io.read(chunk_size):
    sha1.update(buf)
  return sha1.hexdigest()

def sha1sum_dict(obj):
  import json, hashlib
  sha1 = hashlib.sha1()
  sha1.update(json.dumps(obj, sort_keys=True, separators=None, ensure_ascii=True).encode())
  return sha1.hexdigest()

def sanitize_sha1sum(val):
  import re
  m = re.match(r'^[0-9a-f]{40}$', val)
  return val if m else None

def generate_uuid():
  return str(uuid.uuid4())

def sanitize_uuid(val):
  try:
    return str(uuid.UUID(val))
  except:
    return None

def collapse(L):
  if len(L) == 1:
    return L[0]
  else:
    return L

def route_join_with_or_without_slash(blueprint, *routes, **kwargs):
  ''' Like @app.route but doesn't care about trailing slash or not
  '''
  def wrapper(func):
    routes_stripped = join_routes(*routes)
    if routes_stripped:
      blueprint.route(routes_stripped, **kwargs)(func)
    blueprint.route(routes_stripped + '/', **kwargs)(func)
    return func
  return wrapper



import json
import logging
import functools
import urllib.parse
from werkzeug.security import safe_join

logger = logging.getLogger(__name__)

def dict_filter_none(d):
  return { k: v for k, v in d.items() if v }

def join_routes(*routes):
  ''' Utility function for joining routes while striping extraneous slashes
  '''
  return '/' + '/'.join([route.strip('/') for route in routes if route.strip('/')])

def re_full(expr):
  if not expr.startswith('^'): expr = '^' + expr
  if not expr.endswith('$'): expr = expr + '$'
  return expr

def secure_filepath(filepath):
  ''' Ensures this will be a sanitized relative path
  '''
  secured_filepath = safe_join('.', filepath)
  assert secured_filepath is not None, "Filepath wasn't secure"
  secured_filepath = secured_filepath[2:]
  if not secured_filepath:
    logger.debug("Filepath became empty while securing")
    return ''
  return secured_filepath

def secure_url(url):
  parsed = urllib.parse.urlparse(url)
  assert parsed.scheme in {'https', 'http', 'ftp'}, 'Invalid scheme'
  return url

def is_remote(url):
  try:
    secure_url(url)
    return True
  except:
    return False

def importdir(_dirname_, _package_, _globals_):
  import os, importlib
  for f in os.listdir(_dirname_):
    if f.startswith('_') or f.startswith('.'):
      continue
    if f.endswith('.py'):
      modname = f[:-len('.py')]
    elif os.path.isdir(os.path.join(_dirname_, f)):
      modname = f
    else:
      continue
    mod = importlib.import_module('.{}'.format(modname), _package_)
    _globals_.update(**{modname: mod})

def importdir_deep(_dirname_, _package_, _globals_, filter_mod=lambda m, k, v: not k.startswith('_')):
  import os, importlib
  for f in os.listdir(_dirname_):
    if f.startswith('_') or f.startswith('.'):
      continue
    if f.endswith('.py'):
      modname = f[:-len('.py')]
    elif os.path.isdir(os.path.join(_dirname_, f)):
      modname = f
    else:
      continue
    mod = importlib.import_module('.{}'.format(modname), _package_)
    _globals_.update(**{
      k: v
      for k, v in mod.__dict__.items()
      if filter_mod(mod, k, v)
    })

def run_in_executor(f):
  import asyncio
  @functools.wraps(f)
  def inner(*args, **kwargs):
    loop = asyncio.get_running_loop()
    return loop.run_in_executor(None, lambda: f(*args, **kwargs))
  return inner

def exception_as_dict(exc):
  if getattr(exc, 'as_dict', None) is None:
    return dict(cls=exc.__class__.__name__, message=str(exc))
  else:
    return dict(cls=exc.__class__.__name__, **exc.as_dict())



from appyter.cli import cli
from click_default_group import DefaultGroup

@cli.group(cls=DefaultGroup, default='dispatcher', help='Helpers for orchestrating appyters using different runtimes')
def orchestration():
  pass



import os

def url_for(directory, public=False, **kwargs):
  from appyter.context import get_env
  from appyter.util import safe_join, join_routes
  config = get_env()
  url = None
  if config['DEBUG']:
    try:
      from flask import url_for
      modified_directory = '.'.join(('__main__', directory)) if directory == 'static' else directory
      url = url_for(modified_directory, **kwargs)
    except:
      pass
  #
  if url is None:
    filename = kwargs.get('filename', kwargs.get('path'))
    assert filename is not None
    url = safe_join(config['PREFIX'], directory, filename)
  #
  if public:
    try:
      from flask import request
      url = join_routes(request.url_root, url)[1:]
    except:
      pass
  #
  return url



import re
from appyter.fields import Field
from appyter.util import secure_filepath, join_routes, re_full

class FileField(Field):
  ''' Represing a uploadable File and facilitating that file upload.
  The field ends up being a string of the file path, for example:
  ```python
  import pandas as pd
  df = pd.read_csv({{ FileField(name='my-csv', label='My CSV') }})
  ```

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param constraint: A regular expression for validating the file name.
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param default: (str) A default value as an example and for use during prototyping
  :param examples: (Optional[Dict[str, str]]) Named url paths to example files to upload
    paths can be relative i.e. `{ "my_file.txt": url_for('static', filename='my_file.txt') }`, or a remote url.
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Remaining arguments passed down to :class:`appyter.fields.Field`'s constructor.
  '''
  def __init__(self, constraint=r'.*', examples={}, **kwargs):
    super().__init__(
      constraint=constraint,
      examples=examples,
      **kwargs,
    )

  @property
  def raw_value(self):
    if type(self.args['value']) == str:
      return secure_filepath(self.args['value'])
    else:
      return None

  def constraint(self):
    if self.raw_value is None:
      return not self.args.get('required')
    else:
      return re.match(re_full(self.args['constraint']), self.raw_value)

  @property
  def public_url(self):
    try:
      from flask import request
      return join_routes(request.base_url, self.value)[1:]
    except:
      from appyter.context import get_env
      config = get_env()
      return join_routes(config.get('PUBLIC_URL', 'file:///' + config.get('CWD')), self.value)[1:]



''' ```eval_rst
Each submodule represents a profile extending from :mod:`appyter.profiles.default` which provide different rendering themes.
``` '''

import os
from appyter.cli import cli
from appyter.util import importdir

# capture all profiles in the directory and add to profiles/globals
profiles = {}
importdir(os.path.dirname(__file__), __package__, profiles)
globals().update(**profiles)

# register list_profiles command based on profile names and first line of module docstring
@cli.command(help='List the available profiles for the profile option')
def list_profiles():
  profile_name_max_size = max(map(len, profiles.keys()))
  for profile, mod in profiles.items():
    doc_first_line = mod.__doc__.strip().splitlines()[0].strip()
    print(profile.ljust(profile_name_max_size+1), doc_first_line)



import os
import uuid
import click
import logging
logger = logging.getLogger(__name__)

from appyter.cli import cli
from appyter.ext.click import Json, click_option_setenv, click_argument_setenv

def create_app(**kwargs):
  ''' Completely initialize the flask application
  '''
  from aiohttp import web
  from aiohttp_wsgi import WSGIHandler
  from aiohttp_remotes import setup, XForwardedRelaxed
  #
  from flask import Flask, Blueprint, current_app, redirect
  from flask_cors import CORS
  #
  from appyter.render.flask_app.socketio import socketio
  from appyter.render.flask_app.core import core
  import appyter.render.flask_app.static
  import appyter.render.flask_app.download
  import appyter.render.flask_app.execution
  if kwargs['debug']:
    import appyter.render.flask_app.livereload
  #
  from appyter.context import get_env, find_blueprints
  from appyter.util import join_routes
  config = get_env(**kwargs)
  #
  if config['DEBUG']:
    logging.basicConfig(
      level=logging.DEBUG,
      format='%(name)s %(message).80s',
    )
  else:
    logging.basicConfig(
      level=logging.WARNING,
      format='%(name)s %(message).80s',
    )
    logging.getLogger(__package__).setLevel(logging.INFO)
  #
  logger.info('Initializing aiohttp...')
  app = web.Application()
  app['config'] = config
  #
  logger.info('Initializing socketio...')
  socketio.attach(app, join_routes(config['PREFIX'], 'socket.io'))
  #
  logger.info('Initializing flask...')
  flask_app = Flask(__name__, static_url_path=None, static_folder=None)
  CORS(flask_app)
  flask_app.config.update(config)
  flask_app.debug = config['DEBUG']
  #
  logger.info('Registering blueprints...')
  flask_app.register_blueprint(core)
  for blueprint_name, blueprint in find_blueprints(config=flask_app.config).items():
    if isinstance(blueprint, Blueprint):
      flask_app.register_blueprint(blueprint, url_prefix='/'+blueprint_name.strip('/'))
    elif callable(blueprint):
      blueprint(flask_app, url_prefix='/'+blueprint_name.strip('/'))
    else:
      raise Exception('Unrecognized blueprint type: ' + blueprint_name)
  #
  if app['config']['PREFIX'].strip('/'):
    logger.info('Registering prefix redirect')
    async def redirect_to_prefix(request):
      path = request.match_info['path']
      if path == app['config']['PREFIX'].strip('/'): path = ''
      raise web.HTTPFound(join_routes(app['config']['PREFIX'], path) + '/')
    app.router.add_get('/{path:[^/]*}', redirect_to_prefix)
  #
  logger.info('Registering flask with aiohttp...')
  wsgi_handler = WSGIHandler(flask_app)
  app.router.add_route('*', join_routes(app['config']['PREFIX'], '{path_info:.*}'), wsgi_handler)
  if flask_app.config['PROXY']:
    logger.info('Applying proxy fix middleware...')
    import asyncio
    asyncio.get_event_loop().run_until_complete(setup(app, XForwardedRelaxed()))
  return app

# register flask_app with CLI
@cli.command(help='Run a flask web application based on an appyter-enabled jupyter notebook')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_option_setenv('--prefix', envvar='APPYTER_PREFIX', default='/', help='Specify the prefix for which to mount the webserver onto')
@click_option_setenv('--profile', envvar='APPYTER_PROFILE', default='default', help='Specify the profile to use for rendering')
@click_option_setenv('--extras', envvar='APPYTER_EXTRAS', default=[], type=Json(), multiple=True, help='Specify extras flags')
@click_option_setenv('--host', envvar='APPYTER_HOST', default='127.0.0.1', help='The host the flask server should run on')
@click_option_setenv('--socket', envvar='APPYTER_SOCKET', default=None, help='The socket to mount to (when running in production)')
@click_option_setenv('--port', envvar='APPYTER_PORT', type=int, default=5000, help='The port this flask server should run on')
@click_option_setenv('--proxy', envvar='APPYTER_PROXY', type=bool, default=False, help='Whether this is running behind a proxy and the IP should be fixed for CORS')
@click_option_setenv('--data-dir', envvar='APPYTER_DATA_DIR', default='data', help='The directory to store data of executions')
@click_option_setenv('--dispatcher', envvar='APPYTER_DISPATCHER', type=str, help='The URL to the dispatcher, otherwise use an embedded dispatcher')
@click_option_setenv('--dispatcher-url', envvar='APPYTER_DISPATCHER_URL', type=str, help='The URL that the internal dispatcher would contact')
@click_option_setenv('--dispatcher-image', envvar='APPYTER_DISPATCHER_IMAGE', type=str, help='The docker image that the internal dispatcher would use')
@click_option_setenv('--secret-key', envvar='APPYTER_SECRET_KEY', default=str(uuid.uuid4()), help='A secret key for flask')
@click_option_setenv('--debug', envvar='APPYTER_DEBUG', type=bool, default=True, help='Whether or not we should be in debugging mode, not for use in multi-tenant situations')
@click_option_setenv('--static-dir', envvar='APPYTER_STATIC_DIR', default='static', help='The folder whether staticfiles are located')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def flask_app(**kwargs):
  import logging
  if kwargs.get('debug'):
    logging.basicConfig(
      level=logging.DEBUG,
      format='%(name)s %(message).80s',
    )
  else:
    logging.basicConfig(
      level=logging.WARNING,
      format='%(name)s %(message).80s',
    )
    logging.getLogger(__package__).setLevel(logging.INFO)
  #
  if kwargs.get('socket'):
    from aiohttp import web
    socket = kwargs['socket']
    logging.info(f"Launching aiohttp server on {socket}")
    app = create_app(**kwargs)
    if ':' in socket:
      host, port = socket.split(':')
      web.run_app(app, host=host, port=int(port))
    else:
      web.run_app(app, path=socket)
  elif kwargs.get('debug'):
    from appyter.render.flask_app.development import serve
    serve(__file__, **kwargs)
  else:
    from appyter.render.flask_app.production import serve
    serve(__file__, **kwargs)



import os
import click
import nbformat
from appyter.ext.fs import Filesystem
from appyter.cli import cli
from appyter.context import get_env, get_jinja2_env
from appyter.ext.click import click_argument_setenv

@cli.command(help='Clean an appyters output & metadata to minimize unnecessary data transfers in production')
@click_argument_setenv('ipynb', type=str, envvar='APPYTER_IPYNB')
def nbclean(ipynb, **kwargs):
  nb = nbformat.read(open(ipynb, 'r'), as_version=4)
  for cell in nb.cells:
    if cell['cell_type'] == 'code':
      cell['execution_count'] = None
      cell['metadata'] = {}
      cell['outputs'] = []
    elif cell['cell_type'] == 'markdown':
      if 'outputs' in cell:
        del cell['outputs']
      if 'execution_count' in cell:
        del cell['execution_count']
  if 'widgets' in nb['metadata']:
    del nb['metadata']['widgets']
  nbformat.write(nb, open(ipynb, 'w'))



from appyter.fields import Field, FieldConstraintException
from appyter.ext.json import try_json_loads

class MultiChoiceField(Field):
  ''' Represing a multi-selectable combo box.
  The resulting selection is represented with a list of strings that were chosen.

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param choices: (Union[List[str], Set[str], Dict[str, str]]) A set of choices that are available for this field or lookup table mapping from choice label to resulting value
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param default: (float) A default value as an example and for use during prototyping
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Remaining arguments passed down to :class:`appyter.fields.Field`'s constructor.
  '''
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    assert self.choices, 'MultiChoiceField requires choices'

  @property
  def raw_value(self):
    value = try_json_loads(self.args['value'])
    if not value:
      return []
    elif type(value) == list:
      return value
    elif type(value) == str:
      return [value]
    else:
      return [self.args['value']]

  def constraint(self):
    if not self.raw_value:
      return not self.args.get('required')
    else:
      return type(self.raw_value) == list and all(v in self.choices for v in self.raw_value)

  @property
  def value(self):
    if not self.constraint():
      raise FieldConstraintException(
        field=self.field,
        field_name=self.args['name'],
        value=self.raw_value,
      )
    elif type(self.choices) == dict:
      return [self.choices[v] for v in self.raw_value]
    else:
      return self.raw_value
